from grainy import core, const
import unittest

class TestNamespace(unittest.TestCase):

    def test_init(self):
        ns = core.Namespace("a.b.c")
        self.assertEqual(ns.value, "a.b.c")

        ns = core.Namespace("a.b.*")
        self.assertEqual(ns.value, "a.b")

        ns = core.Namespace(["a","b","c"])
        self.assertEqual(ns.value, "a.b.c")

        ns = core.Namespace(["a","b",1])
        self.assertEqual(ns.value, "a.b.1")


    def test_iter(self):
        ns = core.Namespace("a.b.c")
        self.assertEqual([k for k in ns], ["a","b","c"])

    def test_container(self):
        ns = core.Namespace("a.b.c")
        container, tail = ns.container()
        self.assertEqual(container, {"a":{"b":{"c":{}}}})
        self.assertEqual(tail, {})

        container, tail = ns.container({"d":123})
        self.assertEqual(container, {"a":{"b":{"c":{"d":123}}}})
        self.assertEqual(tail, {"d":123})

    def test_match(self):
        ns = core.Namespace("a.b.c")
        self.assertEqual(ns.match(["a","b"]), True)
        self.assertEqual(ns.match(["a"]), True)
        self.assertEqual(ns.match(["a","*"]), True)
        self.assertEqual(ns.match(["a","*","c"]), True) 
        self.assertEqual(ns.match(["a","b","c"]), True)
        self.assertEqual(ns.match(["a","*","c","d"]), False)
        self.assertEqual(ns.match(["a","b","c","d"]), False)
        self.assertEqual(ns.match(["b"]), False)
        self.assertEqual(ns.match(["a","c"]), False)



from builtins import str
from past.builtins import basestring
from builtins import object
import grainy.const as const

def list_key_handler(row, idx):
    if isinstance(row, dict):
        return row.get("id", row.get("name", str(idx)))
    return idx

class Namespace(object):
    """
    Object representing a permissioning namespace
    """

    def __init__(self, value):
        self.set(value)

    def __unicode__(self):
        return self.value

    def __str__(self):
        return self.value

    def __iter__(self):
        for k in self.value.split("."):
            yield k

    def __eq__(self, other):
        return str(self) == str(other)

    def set(self, value):
        if isinstance(value, list):
            value = ".".join([str(v) for v in value])
        if len(value) > 2 and value[-2:] == ".*":
            value = value[:-2]
        self.value = value
        self.keys = [k for k in self]
        self.length = len(self.keys)

    def match(self, keys, partial=True):
        """
        Check if the value of this namespace is matched by
        keys

        '*' is treated as wildcard

        Arguments:

        keys -- list of keys

        Examples:

            ns = Namespace("a.b.c")
            ns.match(["a"]) #True
            ns.match(["a","b"]) #True
            ns.match(["a","b","c"]) #True
            ns.match(["a","*","c"]) #True
            ns.match(["b","b","c"]) #False

        """
        if not partial and len(keys) != self.length:
            return False
        c = 0
        for k in keys:
            if c >= self.length:
                return False
            a = self.keys[c]
            if a != "*" and k != "*" and k != a:
                return False
            c += 1

        return True

    def container(self, data=None):
        """
        Creates a dict built from the keys of this namespace

        Returns the dict created as well as the tail in a tuple

        Example:

            self.value = "a.b.c"

            container, tail = self.container()
            #{"a":{"b":{"c":{}}}, {}

            container, tail = self.container({"d":123})
            #{"a":{"b":{"c":{"d":123}}}, {"d":123}
        """

        if data is None:
            data = {}

        root = p = d = {}
        j = 0
        k = None
        for k in self:
            d[k] = {}
            p = d
            d = d[k]
        if k is not None:
            p[k] = data
        return (root, p[k])


class Permission(object):
    """
    Permission object defined by a namespace and a permission bitmask

    Arguments:

    namespace -- string or Namespace instance defining the namespace 
    value -- permission bitmask
    """

    def __init__(self, namespace, value):
        if isinstance(namespace, basestring):
            namespace = Namespace(namespace)

        self.namespace = namespace
        self.value = value

    def __eq__(self, other):
        r = (
            other.namespace == self.namespace and
            other.value == self.value
        )
        return r

    def has_value(self):
        """
        Check that value has been set
        """
        return (self.value is not None)

    def check(self, level):
        """
        Check if permission flagset contains the specified
        permission level
        """
        if not self.has_value():
            return False
        return (self.value & level) != 0


class PermissionSet(object):
    """
    Holds a set of Namespaces and permissions to run permission checks
    on

    Can also be applied to a data dict to remove keys that are not
    accessible according the permissions in the set

    Keyword arguments:
    rules -- list of Permission objects or dict of namespace(str) : permission(int) pairs
    """

    def __init__(self, rules=None):

        if rules is None:
            rules = []

        self.permissions = {}
        self.index = {}
        self.read_access_map = {}

        if type(rules) == list:
            for permission in rules:
                self.__add__(permission)
        elif type(rules) == dict:
            for ns, p in list(rules.items()):
                self.__add__(Permission(ns, p))

    @property
    def namespaces(self):
        """
        Returns list of all namespaces registered in this permission set
        """
        return list(self.permissions.keys())

    def __iter__(self):
        for v in list(self.permissions.values()):
            yield v

    def __contains__(self, item):
        return item in self.permissions

    def __radd__(self, other):
        self.__add__(other)

    def __add__(self, other):
        if isinstance(other, Permission):
            self.permissions[str(other.namespace)] = other
        self.update_index()

    def __setitem__(self, key, other, reindex=True):
        if isinstance(other, Permission):
            self.permissions[key] = other
        elif isinstance(other, int):
            self.permissions[key] = Permission(key, other)
        else:
            raise TypeError(
                "Value needs to be a Permission instance or a permission flag")
        if reindex:
            self.update_index()

    def __delitem__(self, namespace):
        if namespace in self.permissions:
            del self.permissions[namespace]
        else:
            raise KeyError(
                "No permission registered under namespace '%s'" % namespace)
        self.update_index()

    def update(self, permissions):
        """
        Update the permissionset with a dict of namespace<str>:permission<Permission|int|long>
        pairs

        Example:

            pset.update(
              {
                "a" : const.PERM_READ,
                "b" : Permission("b", const.PERM_RW)
              }
            )
        """
        for k, v in list(permissions.items()):
            self.__setitem__(k, v, reindex=False)
        self.update_index()

    def update_index(self):
        """
        Regenerates the permission index for this set

        Called everytime a rule is added / removed / modified in
        the set
        """

        # update index

        idx = {}
        for _, p in sorted(self.permissions.items(), key=lambda x: str(x[0])):
            branch = idx
            parent_p = const.PERM_DENY
            for k in p.namespace.keys:
                if not k in branch:
                    if k != "*":
                        branch[k] = {"__": parent_p}
                    else:
                        branch[k] = {"__": None}
                    branch[k].update(__implicit=True)

                branch = branch[k]
                parent_p = branch["__"]
            branch["__"] = p.value
            branch["__implicit"] = False

        self.index = idx

        # update read access map

        ramap = {}

        def update_ramap(branch_idx):
            r = {"__": False}
            for k, v in list(branch_idx.items()):
                if k != "__" and k != "__implicit":
                    r[k] = update_ramap(v)

            if branch_idx["__"] is not None and (branch_idx["__"] & const.PERM_READ) != 0:
                r["__"] = True
            return r

        for k, v in list(idx.items()):
            ramap[k] = update_ramap(v)

        self.read_access_map = ramap

        return self.index

    def _check(self, keys, branch, flags=None, i=0, explicit=False, l=0):

        try:
            key = keys[i]
        except IndexError:
            return flags, i

        p = 0
        r = 0
        j = 0
        a = 0

        if not l:
            l = len(keys)

        if key in branch:
            if explicit and branch[key].get("__implicit") and i + 1 >= l:
                p, r = 0, 0
            else:
                p, r = self._check(keys, branch[key], flags=branch[key].get(
                    "__", flags), i=i + 1, explicit=explicit, l=l)
        if "*" in branch:
            if explicit and branch["*"].get("__implicit") and i + 1 >= l:
                j, a = 0, 0
            else:
                j, a = self._check(keys, branch[
                                   "*"], flags=branch["*"].get("__", flags), i=i + 1, explicit=explicit, l=l)

        if explicit and r == 0 and a == 0:
            return 0, i

        if j is not None:
            if r < a or p is None:
                return j, a
        if p is not None:
            if r > i or flags is None:
                return p, r
        return flags, i

    def get_permissions(self, namespace, explicit=False):
        """
        Returns the permissions level for the specified namespace

        Arguments:

        namespace -- permissioning namespace (str)
        explicit -- require explicitly set permissions to the provided namespace

        Returns:

        int -- permissioning flags
        """

        if not isinstance(namespace, Namespace):
            namespace = Namespace(namespace)
        keys = namespace.keys
        p, _ = self._check(keys, self.index, explicit=explicit)
        return p


    def check(self, namespace, level, explicit=False):
        """
        Checks if the permset has permission to the specified namespace
        at the specified level

        Arguments:

        namespace -- permissioning namespace (str) 
        level -- permissioning level (int) (PERM_READ for example)
        explicit -- require explicitly set permissions to the provided namespace

        Returns:

        bool
        """

        return (self.get_permissions(namespace, explicit=explicit) & level) != 0

    def apply(self, data, path=None, applicator=None):
        """
        Apply permissions in this set to the provided data, effectively
        removing all keys from it are not permissioned to be viewed

        Arguments:

        data -- dict of data 

        Returns:

        Cleaned data
        """

        if applicator:
            applicator.pset = self
        else:
            applicator = Applicator(self)

        return applicator.apply(data, path=path)


class Applicator(object):

    handlers = {}

    def __init__(self, pset):
        self.pset = pset

    def handler(self, path, key=None, explicit=False):
        if not isinstance(path, Namespace):
            path = Namespace(path)
        self.handlers[str(path)] = {
            "namespace" : path,
            "key" : key,
            "explicit" : explicit
        }

    def apply(self, data, path=None):
        """
        Apply permissions in this set to the provided data, effectively
        removing all keys from it are not permissioned to be viewed

        Arguments:

        data -- dict of data 

        Returns:

        Cleaned data
        """

        if path is None:
            path = []

        if not isinstance(data, dict):
            return data

        def _enumerate(value):
            if isinstance(value, list):
                for k, v in enumerate(value):
                    yield k, v
            elif isinstance(value, dict):
                for k, v in value.items():
                    yield k, v

        def _set(container, key, value):
            if isinstance(container, list):
                container.append(value)
            else:
                container[key] = value

        def _apply(ramap, value, status=False, wc=False, path=[]):


            if not isinstance(value, dict) and not isinstance(value, list):
                if status:
                    return value
                else:
                    return None

            if not wc:
                status = ramap.get("__", status)

            handler = None
            key_handler = None
            if path and self.handlers:
                namespace = Namespace(path)
                for _handler in self.handlers.values():
                    if namespace.match(_handler.get("namespace").keys, partial=False):
                        handler = _handler
                        key_handler = handler.get("key")
                        break

            if isinstance(value, list):
                if not key_handler:
                    key_handler = list_key_handler
                rv = []
            else:
                rv = {}

            for k, v in _enumerate(value):
                if key_handler:
                    k = key_handler(v, k)
                if isinstance(v, dict) or isinstance(v, list):
                    if k in ramap:
                        r = _apply(ramap[k], v, status=status, path=path+[k])
                        if r:
                            _set(rv, k, r)
                    elif "*" in ramap:
                        r = _apply(ramap["*"], v, status=status, wc=True, path=path+[k])
                        if r:
                            _set(rv, k, r)
                    elif status:
                        _set(rv, k, v)
                else:
                    if k in ramap:
                        if ramap[k].get("__", True):
                            _set(rv, k, v)
                    elif "*" in ramap and ramap["*"].get("__", True):
                        _set(rv, k, v)
                    elif status:
                        _set(rv, k, v)

            return rv

        # loop through all the handlers that specify the `explicit` arguments
        # and temprorarily add deny rules for those to the targeted permissionset
        tmpns = {}
        for ns, handler in self.handlers.items():
            if handler.get("explicit"):
                p = self.pset.get_permissions(ns)
                if p & const.PERM_READ:
                    if ns in self.pset.permissions:
                        continue
                    tmpns[ns] = p
                    self.pset[ns] = const.PERM_DENY

        # apply permissions
        rv = _apply(self.pset.read_access_map, data)

        # remove temporarily added deny rules 
        for ns, p in tmpns.items():
            if p is None:
                del self.pset[ns]
            else:
                self.pset[ns] = p

        return rv





PERM_READ = 0x01
PERM_UPDATE = 0x02
PERM_CREATE = 0x04
PERM_DELETE = 0x08
PERM_WRITE = PERM_UPDATE | PERM_CREATE | PERM_DELETE
PERM_DENY = 0
PERM_RW = PERM_READ | PERM_WRITE

MATCH_NO = 0
MATCH_YES = 1
MATCH_PARTIAL = 2




from setuptools import find_packages, setup


version = open('facsimile/VERSION').read().strip()
requirements = open('facsimile/requirements.txt').read().split("\n")
test_requirements = open('facsimile/requirements-test.txt').read().split("\n")


setup(
    name='grainy',
    version=version,
    author='20C',
    author_email='code@20c.com',
    description='granular permissions utility',
    long_description='',
    license='LICENSE.txt',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],
    packages = find_packages(),
    include_package_data=True,
    url='https://github.com/20c/grainy',
    download_url='https://github.com/20c/grainy/%s' % version,

    install_requires=requirements,
    test_requires=test_requirements,

    zip_safe=True
)



from grainy import core, const
import pytest
import sys
import unittest
import json

performance_test = pytest.mark.skipif(
    not pytest.config.getoption("--performance"),
    reason="need --performance option to run"
)

p1 = core.Permission("a", const.PERM_READ)
p2 = core.Permission("a.b.c", const.PERM_RW)

pdict = {
    "a" : const.PERM_READ,
    "a.b.c" : const.PERM_RW,
    "a.b.*.d" : const.PERM_DENY,
    "a.c" : const.PERM_WRITE,
    "b.c" : const.PERM_READ
}


pdict2 = {
    "a" : const.PERM_READ,
    "a.b.c" : const.PERM_RW,
    "a.b.e" : const.PERM_DENY,
    "a.b.*.d" : const.PERM_DENY,
    "e.f" : const.PERM_READ,
    "e.*.g" : const.PERM_WRITE,
    "e.*.g.a" : const.PERM_READ,
    "e.*.g.b" : const.PERM_RW,
    "e.h.g" : const.PERM_DENY,
    "f.g" : const.PERM_READ
}

pdict3 = {
    "a.100" : const.PERM_READ,
    "a.b.c" : const.PERM_RW,
    "a.b.e" : const.PERM_DENY,
    "a.b.*.d" : const.PERM_DENY,
    "e.f" : const.PERM_READ,
    "e.*.g" : const.PERM_WRITE,
    "e.*.g.a" : const.PERM_READ,
    "e.*.g.b" : const.PERM_RW,
    "e.h.g" : const.PERM_DENY
}

pdict4 = {
    "a.b" : const.PERM_READ,
    "x" : const.PERM_READ,
    "x.z" : const.PERM_DENY,
    "nested.*.data.public" : const.PERM_READ
}


class TestPermissionSet(unittest.TestCase):

    def test_init(self):
        pset = core.PermissionSet([p1,p2])
        self.assertEqual(pset.permissions["a"], p1)
        self.assertEqual(pset.permissions["a.b.c"], p2)

        pset = core.PermissionSet(pdict)
        self.assertEqual(pset.permissions["a"], p1)
        self.assertEqual(pset.permissions["a.b.c"], p2)

    def test_update_index(self):
        pset = core.PermissionSet(pdict)
        expected = {
            'a': {
                '__': 1,
                '__implicit' : False,
                'c': {
                    '__implicit' : False,
                    '__': 14
                },
                'b': {
                    '__': 1,
                    '__implicit' : True,
                    'c': {
                        '__': 15,
                        '__implicit' : False
                    },
                    '*': {
                        '__implicit' : True,
                        '__': None,
                        'd': {
                            '__implicit' : False,
                            '__': 0
                        }
                    }
                }
            },
            'b': {
                '__implicit' : True,
                '__': 0,
                'c': {
                    '__implicit' : False,
                    '__': 1
                }
            }
        }

        self.assertEqual(pset.index, expected)

    def test_contains(self):
        pset = core.PermissionSet(pdict)
        self.assertIn("a", pset)
        self.assertIn("a.b.c", pset)
        self.assertNotIn("x", pset)

    def test_update(self):

        pset = core.PermissionSet(pdict)

        pset.update({"x":const.PERM_READ, "z":core.Permission("z", const.PERM_READ)})
        self.assertIn("a", pset)
        self.assertIn("a.b.c", pset)
        self.assertIn("x", pset)
        self.assertIn("z", pset)

        self.assertEqual(pset.check("x", const.PERM_READ), True)
        self.assertEqual(pset.check("z", const.PERM_READ), True)


    def test_setitem_delitem(self):
        pset = core.PermissionSet()
        pset["a"] = const.PERM_READ
        pset["a.b"] = const.PERM_RW
        pset["b"] = const.PERM_READ

        self.assertEqual(pset.permissions["a"].check(const.PERM_READ), True)
        self.assertEqual(pset.permissions["a.b"].check(const.PERM_WRITE), True)
        self.assertEqual(pset.permissions["b"].check(const.PERM_READ), True)

        pset["a.b"] = const.PERM_READ

        self.assertEqual(pset.permissions["a.b"].check(const.PERM_WRITE), False)

        del pset["b"]

        self.assertNotIn("b", pset)


    def test_check(self):
        pset = core.PermissionSet(pdict2)

        self.assertEqual(pset.check("a.b", const.PERM_READ), True)
        self.assertEqual(pset.check("a.b.c", const.PERM_WRITE), True)
        self.assertEqual(pset.check("a.b.d", const.PERM_READ), True)
        self.assertEqual(pset.check("a.b.c.d", const.PERM_READ), False)
        self.assertEqual(pset.check("e.f", const.PERM_READ), True)
        self.assertEqual(pset.check("e", const.PERM_READ), False )
        self.assertEqual(pset.check("e.j.g", const.PERM_WRITE), True)
        self.assertEqual(pset.check("e.k.g.a", const.PERM_WRITE), False)
        self.assertEqual(pset.check("e.h.g", const.PERM_READ), False)
        self.assertEqual(pset.check("e.h.g.a", const.PERM_WRITE), False)
        self.assertEqual(pset.check("e.m.g.a", const.PERM_WRITE), False)
        self.assertEqual(pset.check("e.m.g.b", const.PERM_RW), True)
        self.assertEqual(pset.check("f", const.PERM_WRITE), False)
        self.assertEqual(pset.check("f.g", const.PERM_READ), True)

    def test_check_explicit(self):
        pset = core.PermissionSet(pdict)
        self.assertEqual(pset.check("a.b", const.PERM_READ, explicit=True), False)
        self.assertEqual(pset.check("a", const.PERM_READ, explicit=True), True )
        self.assertEqual(pset.check("a", const.PERM_WRITE, explicit=True), False)
        self.assertEqual(pset.check("a.b.c", const.PERM_WRITE, explicit=True), True)
        self.assertEqual(pset.check("a.b.c", const.PERM_READ, explicit=True), True)

    def test_apply(self):
        pset = core.PermissionSet(pdict2)
        data = {
            "a" : {
                "b" : {
                    "c" : {
                        "A" : True
                    },
                    "d" : {
                        "A" : True
                    },
                    "e" : {
                        "A" : False
                    }
                }
            },
            "f": {
                "g" : True
            }
        }

        expected = {
            "a" : {
                "b" : {
                    "c" : {
                        "A" : True
                    },
                    "d" : {
                        "A" : True
                    }
                }
            },
            "f": {
                "g" : True
            }
        }

        rv = pset.apply(data)
        self.assertEqual(rv, expected)

    def test_apply_explicit(self):
#pdict = {
#    "a" : const.PERM_READ,
#    "a.b.c" : const.PERM_RW,
#    "a.b.*.d" : const.PERM_DENY,
#    "a.c" : const.PERM_WRITE,
#    "b.c" : const.PERM_READ
#}
        pset = core.PermissionSet(pdict)

        data = {
            "a" : {
                "b": {
                    "c": True,
                    "d": False,
                    "e": True,
                    "f" : { "something" : "else"}
                }
            }
        }

        expected = {
            "a" : {
                "b" : {
                    "c": True,
                    "e": True
                }
            }
        }

        applicator = core.Applicator(pset)
        applicator.handler("a.b.d", explicit=True)
        applicator.handler("a.b.f", explicit=True)
        rv = pset.apply(data, applicator=applicator)
        self.assertEqual(rv, expected)

        expected = {
            "a" : {
                "b" : {
                    "c": True,
                    "d": False,
                    "e": True,
                    "f" : { "something" : "else" }
                }
            }
        }

        print("vla")

        pset["a.b.d"] = const.PERM_READ
        pset["a.b.f"] = const.PERM_READ
        rv = pset.apply(data, applicator=applicator)
        self.assertEqual(rv, expected)



    def test_apply_nested_lists(self):
        pset = core.PermissionSet(pdict4)
        data= {
            "a" : [
                { "id" : "b" },
                { "id" : "c" }
            ],
            "x" : [
                { "custom" : "y" },
                { "custom" : "z" }
            ],
            "nested" : [
                {
                    "data" : [
                        {
                            "level" : "public",
                            "some" : "data",
                            "explicit" : {
                                "sekret" : "data"
                            }
                        },
                        {
                            "level" : "private",
                            "sekret" : "data"
                        }
                    ]
                }
            ]
        }

        expected = {
            "a" : [
                { "id" : "b" }
            ],
            "nested" : [
                {
                    "data" : [
                        { "level" : "public", "some" : "data" }
                    ]
                }
            ],
            "x" : [
                { "custom" : "y" }
            ]
        }

        applicator = core.Applicator(pset)
        applicator.handler("x", key=lambda row,idx: row["custom"])
        applicator.handler("nested.*.data", key=lambda row,idx: row["level"])
        applicator.handler("nested.*.data.public.explicit", explicit=True)
        rv = pset.apply(data, applicator=applicator)
        self.assertEqual(rv, expected)



    @performance_test
    def test_performance(self):

        def mkdataset(depth=3):
            depth = depth - 1
            if depth <= 0:
                return
            return dict([(str(k),mkdataset(depth=depth)) for k in range(1,1000)])
        data = {
            "a" : mkdataset(3),
            "b" : mkdataset(3)
        }

        pset = core.PermissionSet(pdict3)

        import time

        t= time.time()
        cleaned = pset.apply(data)
        diff = time.time() - t

        self.assertLess(diff, 0.002)




from grainy import core, const
import unittest

test_ns = core.Namespace("a.b.c")

class TestPermission(unittest.TestCase):
    
    def test_init(self):
        perm = core.Permission(test_ns, const.PERM_RW)
        self.assertEqual(perm.value, const.PERM_RW)
        self.assertEqual(perm.namespace, test_ns)

        perm = core.Permission("a.b.c", const.PERM_RW)
        self.assertEqual(isinstance(perm.namespace, core.Namespace), True)

    def test_has_value(self):
        perm = core.Permission(test_ns, const.PERM_RW)
        self.assertEqual(perm.has_value(), True)

        perm = core.Permission(test_ns, const.PERM_DENY)
        self.assertEqual(perm.has_value(), True)

        perm = core.Permission(test_ns, None)
        self.assertEqual(perm.has_value(), False)

    def test_check(self):
        perm = core.Permission(test_ns, const.PERM_RW)
        self.assertEqual(perm.check(const.PERM_READ), True)
        self.assertEqual(perm.check(const.PERM_WRITE), True)

        perm = core.Permission(test_ns, const.PERM_READ)
        self.assertEqual(perm.check(const.PERM_READ), True)
        self.assertEqual(perm.check(const.PERM_WRITE), False)

        perm = core.Permission(test_ns, const.PERM_DENY)
        self.assertEqual(perm.check(const.PERM_READ), False)
        self.assertEqual(perm.check(const.PERM_WRITE), False)
